//
//*****************************************************************************
//*                                                                           *
//* COPYRIGHT (c) Maelzel, Inc.,                                              *
//*                                                                           *
//*          Redistribution and use of this code and the algorithms herein    *
//*          contained, with or without modification, is permitted provided   *
//*          that the following conditions are met:                           *
//*          * Redistribution of this code and the algorithms herein          *
//*            contained must retain the above copyright notice, this list    *
//*            of conditions, and the following disclaimer.                   *
//*          * Any use of this code and the algorithms herein contained as    *
//*            text or in binary form must attribute, in a reasonable         *
//*            manner, the provenance and authorship of this code as          *
//*            shown in the copyright above.                                  *
//*                                                                           *
//* DISCLAIMER This code and the algorithms herein contained are provided     *
//*            by the developer ''as is'' without any expressed or implied    *
//*            warranties, including, but not limited to, fitness for any     *
//*            particular purpose or any consequential damage that may        *
//*            occur as a consequence of use or association with this code    *
//*            or the algorithms herein contained.                            *
//*                                                                           *
//* TOP      Clans.java                                                       *
//* CLASS    LifeExperiences.java. (the daily lives of the population)        *
//* AUTHOR   Ovi Chris Rouly                                                  *
//*                                                                           *
//* DATE     07/01/2015                                                       *
//* REVISION X.0.1.4                                                          *
//* BYTES    00,000                                                           *
//* TIME     6:30                                                             *
//* PURPOSE  This object simulates a day in the life of an HG agent.          *
//* METHOD   Each instance of a class is a stand-alone code object.           *
//* DETAILS  A Java document exists that outlines class operation.  Other     *
//*          documentation in the form of programmer notes apply.             *
//* ENVIRON  Developed on Intel Core-2 Quad Q8200 2.33 GHz CPU, Nvidia-based  *
//*          PCI-X graphics acceleration.  The OS was MS Windows 7 Ultimate.  *
//*          Java IDE was JGrasp version 2.0.0_08.  The Java compiler was     *
//*          version 1.8.0.  Verification tests were run in Java JRE          *
//*          build 1.8.0-b132 with client VM build 14.3-b01.                  *
//*                                                                           *
//*****************************************************************************
//

package root.events;                                  // sub-directory storage.

import java.util.ArrayList;                                 //   Java language.
import java.util.Random;                                    //           ditto.
import java.awt.Point;                                      //           ditto.

import root.events.*;                                      // program routines.
import root.hominini.Hominid;                              //            ditto.
import root.ecology.Ecology;                               //            ditto.
import root.ModelConstraints;                              //            ditto.

//
//*****************************************************************************
// Life Experiences

public class LifeExperiences extends ModelConstraints
{

  //
  //***************************************************************************
  // Class constants, variables, and objects.

  // Constants.

  // Variables.
  private static long rng_Seed = 0L;            // this is the GLOBAL RNG seed.
  private static int hLimit = 0;                    // horizontal grid surface.
  private static int vLimit = 0;                      // vertical grid surface.

  private int incestCount = 0;           // runtime count of incest occurances.

  // Objects.
  private static Ecology ecoSys = null; // the virtual space the people occupy.
  private static Random RNG = null;          // pointer to standard RNG object.

  //***************************************************************************
  /** This constructor instantiates methods associated with emulating selected
      phases of life and death among sythetic agents.*/

  public LifeExperiences (long newSeed, int horizontalLimit, int verticalLimit)
  {

    this.rng_Seed = newSeed;                       // copy the GLOBAL RNG seed.
    RNG = new Random(newSeed);          // instantiate an RNG object with seed.

    this.hLimit = horizontalLimit;                        // take a local copy.
    this.vLimit = verticalLimit;                          // take a local copy.

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** WARNING: this is a long (extensive) method.

      This method is a sequence of updating actions that together simulate
      a small part of the daily life of an adult (and juveniles by default)
      in the activities of mating, eating, drinking, bio-physiological
      (metabolic)costs, individual discrete behaviors, and doForage  The
      order of activity is:

      PHYSIOLOGY:  simulate physiology for (B), (J), and (A).
      PSYCHO-SOCIAL SPATIAL INTERACTIONS:  simulate bio-psychology.
      COGNITION: life experiences that change behavior.
      PREDATION: predators take a toll on an individual.
      DROWNING:  tests for and simulates a hominid experiencing a drowning.*/

  public void adultLife(Hominid hominid, Ecology ecoSys, int epoch)
  {

    int h = hominid.getCurrentLoc().x;             // the current local cell X.
    int v = hominid.getCurrentLoc().y;             // the current local cell Y.

    // ------------------------------------------------------------------------
    // PHYSIOLOGY:  simulate biophysiology for (B), (J), and (A).

    float ambientT = ecoSys.getBiomeTemp(h, v);         // for hominidal needs.
    hominid.physiology (ecoSys, ambientT, epoch);

    // ------------------------------------------------------------------------
    // PSYCHO-SOCIAL SPATIAL INTERACTIONS:  simulate bio-psychology.

    socialPsychology (hominid, ecoSys, h, v, epoch); // act within your cohort.

    // ------------------------------------------------------------------------
    // COGNITION: life experiences that change behavior.  Provide homi with
    // recognition of a particular stimulus rather than a general class of
    // stimuli.

    hominid.cognition (ecoSys, "Prey");          // choose the "Prey" stimulus.

    // ------------------------------------------------------------------------
    // PREDATION:  predators take a toll on an individual.

    hominid = predation (hominid); // this operates on all life phases equally.

    // ------------------------------------------------------------------------
    // DROWNING: tests for and execute hominid death by drowning.

    hominid = drowning(hominid, ecoSys, epoch);

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method handles baby-phase nutrition concerns.

      This method is a sequence of updating actions that together simulate
      a small part of the daily life of an baby.  Activities include:

      METABOLIC BIO-PHYSIOLOGY:  simulate biophysiology for (B).
      PSYCHO-SOCIAL SPATIAL INTERACTIONS:  simulate bio-psychology.
      MOVEMENT:  mom handles all of baby's (preweaning) travel needs.
      PREDATION:  predators take a toll on an individual.*/

  public void babyLife(Hominid hominid, Ecology ecoSys, int epoch)
  {

    int h = hominid.getCurrentLoc().x;             // the current local cell X.
    int v = hominid.getCurrentLoc().y;             // the current local cell Y.

    // ------------------------------------------------------------------------
    // METABOLIC BIO-PHYSIOLOGY:  simulate biophysiology for baby.  Mom
    // handles food, water, and baby's ambient temperature needs.  But,
    // while in hominid.bioPhysiology, baby's LifeStatus will be updated
    // and that may precipitate a change from "B" to "J" if this epoch is
    // also baby's birthday.
    hominid.physiology (ecoSys, 0.0F, epoch);

    // ------------------------------------------------------------------------
    // PSYCHO-SOCIAL SPATIAL INTERACTIONS:  simulate bio-psychology.

    socialPsychology (hominid, ecoSys, h, v, epoch); // act within your cohort.

    // ------------------------------------------------------------------------
    // MOVEMENT:  mom handles all of baby's (preweaning) travel needs..

    // ------------------------------------------------------------------------
    // PREDATION:  predators take a toll on an individual.

    hominid = predation (hominid); // this operates on all life phases equally.

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method takes a food gift from a benefactor, assigns it to a
      giftee, and updates the giftee's benefactor property with a pointer
      to the benefactor.*/

  public void benefitTransfer (Hominid benefactor, Hominid giftee, float gift)
  {

    giftee.setBenefactor(benefactor); // tell the giftee who is its benefactor.
    String sex = giftee.bio.getSex();   // used to store food correctly by sex.
    giftee.kit.updateKitFood (sex, gift);    // sex-steered update to kit food.

    // DIAGNOSTIC
    // System.out.println("LifeExperiences.benefitTransfer -" +
    //                    " benefactor sex: " + benefactor.bio.getSex() +
    //                    " giftee sex: " + giftee.bio.getSex() +
    //                    " gift in lbs: " + gift);
    // DIAGNOSTIC

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method simulates the hominid experiencing a sudden, mortally
      devastating event or similar calamity.*/

  public boolean catastrophicDeath(Hominid hominid)
  {

    boolean dead = false;                    // assume the hominid is not dead.

    if (hominid.kit.getPerilStatus() >= hominid.kit.getINDIV_IS_IMPERILED())
    {
      hominid.bio.setHominidLifeStatus("D");                  // set them dead.
      hominid.setExtant(false);                              // reduce extancy.
      hominid.setCauseOfDeath("C");         // record death by catastrophy (C).
      dead = true;                               // set the terminal dead flag.
    }

    //System.out.println("in LifeExperiences.catastropy hominid must die");
    return dead;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method inpects a list of unknown size looking for agents whose
      sex and age parameters make them candidates for a sublist.  It
      returns an integer count of those agents that meet the criteria.
      Assume the source list size is G.T. zero, i.e., that there exists a
      pool of genetic donors through which to search.*/

  public int countBySex (ArrayList<Hominid> cohort, String sex,
                         int minAge)
  {

    int limit = cohort.size();                      // get size of source list.

    int nrToBeReturned = 0;                                 // running counter.
    Hominid p = null;                                      // a test candidate.
    int i = 0;                                  // the count of tested samples.
    do
    {

      p = cohort.get(i);                                        // take a copy.
      if (p.bio.getSex() == sex)                            // compare for sex.
        if (p.bio.getAge() >= minAge)                         // check the age.
          nrToBeReturned++;                             // increment the total.

    } while (++i < limit);

    return nrToBeReturned;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method divides the life phases of each agent into the
      classifications categories of: baby, juvenile, and adult.
      Death (D) marked agents and error modes are discarded.*/

  public void dailyLife (Hominid hominid, Ecology ecoSys, int epoch)
  {

    // Adults (A) and juveniles (J) interact in cohort-specific life.
    if ((hominid.bio.getHominidLifeStatus().equals("A")) ||
        (hominid.bio.getHominidLifeStatus().equals("J")))
      adultLife (hominid, ecoSys, epoch);

    // (B) babies - dependent on adults for nutrition and transport.
    else if (hominid.bio.getHominidLifeStatus().equals("B"))
      babyLife (hominid, ecoSys, epoch);

    // (D) dead - likely caused by orphaned status.
    else if (hominid.bio.getHominidLifeStatus().equals("D"))
    {
      ; // do nothing, Grim Reaper will handle it.
    }

    else
    {
      System.out.println("ERROR: LifeExperiences.dailyLife bad data.");
      System.exit(0);                       // error mode in group life phases.
    }

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method is the equivalent of the hominid foraging for food.*/

  public Hominid doForageFood (Hominid hominid, Ecology ecoSys, int h, int v)
  {

    String localPreyName = "";             // the prey named on the local cell.
    float localProb = 0.0F;                       // local probability of prey.
    float localPounds = 0.0F;                          // pounds of local prey.
    float foodToKit = 0.0F;           // the potential food resource in pounds.

    localPreyName = ecoSys.getForageOption(h, v);// name of available resource.
    localProb = ecoSys.takeForageable(h, v);         // probability of success.
    localPounds = ecoSys.getForagedMass(localPreyName);  // lbs of forageables.
    foodToKit = localProb * localPounds;           // a reduced portion in lbs.

    String sex = hominid.bio.getSex();         // used to store food correctly.
    hominid.kit.updateKitFood (sex, foodToKit);     // update kit food portion.

    return hominid;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method is the equivalent of the hominid foraging for water.*/

  public Hominid doForageWater (Hominid hominid, Ecology ecoSys, int h, int v)
  {

    float waterToKit = 0.0F;           // the potential water resource in gals.

    float qtyNeeded = hominid.bio.getWaterRequirement(); // homis need in gals.

    waterToKit = ecoSys.requestWater (h, v, qtyNeeded);    // attempt to drink.

    hominid.kit.updateKitWater (waterToKit);       // update kit water portion.

    return hominid;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method is the equivalent of the hominid foraging for supplies.*/

  public Hominid doForageKit (Hominid hominid, Ecology ecoSys, int h, int v)
  {


    float supplyStash = 0.0F;                // supplies kept with the hominid.
    float suppliesToKit = 0.0F;             // supplies transferred to the kit.

    supplyStash = hominid.kit.getSuppliesOnHand();// supply in hominid's cache.
    suppliesToKit = supplyStash + (RNG.nextFloat()/10.0F);  // transfer to kit.

    hominid.kit.updateKitSupplies (suppliesToKit); //update kit.

    return hominid;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method explicitly processes a hominid's request for relocation.
      Because a hominid requesting relocation may have been marked as needing
      reassignment (N) well before arriving here, a test for this lifeStatus
      must be undertaken before doing anything else.  All otherwise legitimate
      hominids will be relocated.  If the hominid requesting relocation is also
      female and has an offspring (in an age-specific category) then the
      child too must also be relocated with its mother.  Else, the offspring
      will be left unrelocated.*/

  public void doRelocate (Hominid hominid, Ecology ecoSys, int epoch)
  {

    if ( isIndividualSociallyViable( hominid ) )
    {

      int oldX = hominid.getCurrentLoc().x;    // hominid's current X location.
      int oldY = hominid.getCurrentLoc().y;    // hominid's current Y location.
      int newX = hominid.hal.getReqX(); // hominid's choice of relocation in X.
      int newY = hominid.hal.getReqY(); // hominid's choice of relocation in Y.
      int count = 0;                  // the count of hominid in each location.

      // DIAGNOSTIC.
      //doRelocateDetails (hominid, epoch, hLocX, hLocY, newX, newY);
      // DIAGNOSTIC.

      String lifePhase = hominid.bio.getHominidLifeStatus();     // B, J, or A.

      if (lifePhase.equals("A") || lifePhase.equals("J"))
      {
        hominid.setCurrentLoc(new Point(newX, newY)); // move homi to new cell.
        ecoSys.removeCohortIndividual (oldX, oldY, hominid); //remove from old.
        ecoSys.addCohortIndividual (newX, newY, hominid);         //add to new.

        // if this is a female (F) with children.
        if (hominid.bio.getSex().equals("F") && (hominid.kin.hasChildren()))
        {
          Hominid c = hominid.kin.getLastChild();      // get mom's last child.

          // test if this a Baby and whether or not this is its birthday...
          if ( c.bio.getHominidLifeStatus().equals("B") &&
              ( (epoch - c.getBirthDate() % YEAR) == 0 ) )
          {
            // DIAGNOSTIC.
            //int cLocX = c.getCurrentLoc().x;   // child's current X location.
            //int cLocY = c.getCurrentLoc().y;   // child's current Y location.
            //doRelocateDetails (c, epoch, cLocX, cLocY, newX, newY);
            // DIAGNOSTIC.
            c.setCurrentLoc(new Point(newX, newY));    // move "c" to new cell.
            // remove from old cell cohort list and add to them to new one.
            ecoSys.removeCohortIndividual (oldX, oldY, c);
            ecoSys.addCohortIndividual (newX, newY, c);
          }
        }
      }

    }

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method is a DIAGNOSTIC helper to doRelocate.  It prints to the
      console the relocating agent's ID, sex, age group, and old-new X-Ys.*/

  public void doRelocateDetails (Hominid hominid, int epoch,
                                 int oldX, int oldY, int newX, int newY)
  {

    String sex = "";
    if (hominid.bio.getSex().equals("M"))
      sex = "M";
    else
      sex = "F";

    String id = Long.toString(hominid.getAgentID());
    String ageGroup = hominid.bio.getHominidLifeStatus();

    System.out.println("On day " + epoch + " agent " + id + ", " +
                        sex + " and " + ageGroup + "  moved from X=" +
                   oldX + ", Y=" + oldY + " to X=" + newX + ", Y=" + newY);

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method directs the update of daily life events for each individual
      in the simulation.  The assumed interval of refresh, or Epoch, is
      the day.*/

  public void doSimLife (int epoch, Ecology ecoSys, ArrayList <Hominid> pop)
  {

    // generate an activation list using the current population list.
    int popCount = pop.size();            // count of people in the eco-system.
    int [] uRDist = null;         // a Uniform Random distribution of integers.
    uRDist = getUniformRandom(popCount);           // generate activation list.

    // In uniform-random order, activate (process) one individual at a time.
    // This is an individual-based model simulating discrete-event among
    // a self-organizing, self-scaling population of simulated Hominid agents.
    Hominid hominid = null;
    for (int i = 0; i < popCount; i++)          // check each cell in the list.
    {

      hominid = pop.get(uRDist[i]);         // an individual synthetic hominid.

      // -------------------- Process One Hominid------------- ----------------
      // if not killed by catastrophy or drowning then proceed.
      if (!catastrophicDeath(hominid))
      {

        dailyLife(hominid, ecoSys, epoch);   // one at a time process everyone.
        terminateOrphanInfants (hominid);        // terminate orphaned infants.

      }
      // ------------------- uses the Hominid's migration flag ----------------

    }

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method tests for and terminates the hominid if it is experiencing
      a panic related drowning.  See Hominid.reduceDrives.*/

  public Hominid drowning(Hominid hominid, Ecology ecoSys, int epoch)
  {

    if ( ecoSys.getBiomeName(hominid.getCurrentLoc().x,
                             hominid.getCurrentLoc().y).equals("water") )
    {
      if ((hominid.getWentInWater() + water_days)== epoch)
      {
        hominid.bio.setHominidLifeStatus("D");                // set them dead.
        hominid.setExtant(false);                            // reduce extancy.
        hominid.setCauseOfDeath("S");          // record death by drowning (S).
      }

      hominid.setWentInWater(epoch);                    // set event in memory.

    }
    else
      hominid.setWentInWater(-9); // you're not in water so reset event memory.

    return hominid;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method returns the current count of incest instances.*/

  public int getIncestCount ()
  {

    return this.incestCount;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method generates and returns a set of pseudo-random, uniformly
      distributed values.*/

  public int [] getUniformRandom (int galoisFieldSize)
  {

    final byte EMPTY = (byte) 0x00;                        // symbol for EMPTY.
    final byte FULL =  (byte) 0x01;                         // symbol for FULL.

    int [] uniformRandomList = new int[galoisFieldSize];   // the return field.
    byte [] checkList = new byte [galoisFieldSize];       // used for checking.

    for (int z = 0; z < galoisFieldSize; z++)
      checkList [z] = EMPTY;     // fill the backstop list with NOT USED flags.

    int sample = 0;                               // this is the random sample.
    int g = 0;
    do
    { // continue looping until every "g" in the Galois set is tested.

      sample = RNG.nextInt (galoisFieldSize);
      if (checkList [sample] == EMPTY)
      {
        uniformRandomList [g++] = sample;
        checkList [sample] = FULL;
      }

    } while (g < galoisFieldSize);
    checkList = null;                 // return memory to the operating system.

    return uniformRandomList;                // return the uniform random list.

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method inpects a list of unknown size looking for agents whose
      lifeStatus parameters show them currently within the cohort cell.  It
      returns a Boolean.  No assumption is made about the list size.*/

  public boolean isCohortSociallyViable (ArrayList <Hominid> cohort)
  {

    int limit = cohort.size();                      // get size of source list.
    int count = 0;        // the count of socially viable Hominids in the cell.

    Hominid p = null;                                      // a test candidate.
    for (int i = 0; i < limit; i++)                    // scan the entire list.
    {
      p = cohort.get(i);                                        // take a copy.
      if (isIndividualSociallyViable (p))             // check this individual.
        count++;                                   // then increment the total.
    }

    if (count == 0)
      return false;// cell may be completely empty or just not socially viable.
    else
      return true;    // cell has one or more socially viable agents within it.

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method inpects an individual's lifeStatus parameters and returns
      a Boolean that says if they (individually) are socially viable or not.*/

  public boolean isIndividualSociallyViable (Hominid p)
  {

    return (!p.bio.getHominidLifeStatus().equals("D") &&     // check (D) dead.
            !p.bio.getHominidLifeStatus().equals("B"));      // check (B) baby.

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method inpects a list of unknown size looking for agents whose
      life-status, gender, and age make them candidates for a sublist by
      type.  Assume the source list size is G.T. zero, i.e., that
      there exists a candidate pool over which to search for a suitably
      aged and gendered agent.  The returned product is the set of indices
      pointing to the agents in the source list that meet the requirements.*/

  public ArrayList<Integer> makeListByType (ArrayList<Hominid> cohort,
                                            String sex, int minAge)
  {

    ArrayList<Integer> agentGroup = new ArrayList<Integer>();   // target list.

    boolean dead = false;                        // is candidate actually dead.
    Hominid p = null;                                    // temporary variable.
    int limit = cohort.size();                      // get size of source list.
    int i = 0;                     // an index and the count of tested samples.

    do
    {

      p = cohort.get(i);                                  // use for its tools.
      dead = ( (p.getLifeStatusReport().equals("D")) ? true : false);

      if (!dead && p.bio.getSex().equals(sex))
      {
        if (p.bio.getAge() >= minAge)  // among those, compare for minimum age.
          agentGroup.add(new Integer(i));   // indices of eligible individuals.
      }

    } while (++i < limit);

    return agentGroup;    // this is a grouping taken from the cohort supplied.

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method emulates the mechanics of final mate selection and mating.*/

  public void mating (Hominid hominid, ArrayList <Hominid> cohort, int epoch)
  {

    // using a list of agents in the same spatial cohort look for males of
    // breeding age and kin.  If one or more are found pick one from among
    // them and breed it with an eligible female. If none are found then,
    // do nothing.

    int minAge = hominid.bio.getMaleFertAge();   // minimum male fertility age.
    int availableMales = 0;// this will be the number of sexually mature males.
    ArrayList <Integer> breeders = null;  // list of indices to eligible males.

    if (isCohortSociallyViable (cohort)) // check if cohort is socially viable.
      availableMales = countBySex(cohort, "M", minAge);      // count of males.
    else
      availableMales = 0;                                    // count of males.

    // OK, conception can possibly, occur.
    if (availableMales > 0)
    {
      // an initial list of indices that, in-cohort, may be eligible mates.
      breeders = makeListByType (cohort, "M", minAge);
      int listLength = breeders.size();    // initial length of the mates list.

      // for all sons of hominid, as may exist in the cohort, exclude incest.
      boolean cK = false;   // Murdoch constellation = 1 paterlineal knowledge.
      boolean uK = false;     // Murdoch constellation = 1 no father knowledge.
      boolean wE = false;   // CONSTANT-based comparison of cohort familiarity.
      int male = 0;                     // a pointer into the array of indices.

      // winnow (possible) list further using kinship prohibition filtering.
      if (listLength > 0)
      {
        Hominid possibleMate = null; // a possible male mate to this F hominid.
        while ( ( male < listLength ) && ( listLength > 0) )
        {
          // REM, breeders is a list of indices to possible breeder males
          //      currently in the spatial cohort where hominid is a female.
          possibleMate = cohort.get (breeders.get (male).intValue());

          //cK = hominid.kin.uterineKin (possibleMate);   // consanguineal kin.
          uK = hominid.kin.uterineKin (possibleMate);           // uterine kin.
          wE = hominid.kin.westermarckEffect (possibleMate, epoch); //Wstrmrck.

          // this test can REMOVE many "would-be" incestuous partners.
          if (uK || wE)
            breeders.remove ( male );  //male is Integer pointer into the list.
          else
            male++;

          listLength = breeders.size(); // update the length of the mates list.
        }
        possibleMate = null;          // return memory to the operating system.
      }

      // what remain now are 0 or more candidate males.
      if (listLength > 0)                // if there exists a breedable cohort.
      {

        // randomly select one index that will be bred with the hominid "F".
        int fatherIndex = breeders.get(RNG.nextInt(breeders.size()));
        // add baby to the group.
        Hominid father = cohort.get(fatherIndex);


        // DIAGNOSTIC
        cK = hominid.kin.consanguinalKin (father);
        uK = hominid.kin.uterineKin (father);
        if (cK)
        {
          this.incestCount++;
        }
        if (cK) //(cK || uK)
        {
          System.out.println("LifeExperiences.mating() epoch of conception: " +
                             epoch);
          System.out.println("Male: " + father.getAgentID() +
                             " Female: " + hominid.getAgentID());
          System.out.println("cK: " + cK);
          System.out.println("uK: " + uK);
          System.out.println("Murdoch incest occurred where the mother is\n" +
          "      sibling: " + hominid.kin.isSiblingOf(father) + "\n" +
          "      parent: " + hominid.kin.isParentOf(father) + "\n" +
          "      offspring: " + hominid.kin.isOffspringOf(father) + "\n" +
          "of the father.\n" +
          "Reports follow first mother then father.");
          hominid.report(epoch);
          father.report(epoch);
        }
        // DIAGNOSTIC

        hominid.bio.setMObject(father);                    // the father to be.
        hominid.bio.setPregnancyStatus(true);                  // set the flag.
        hominid.bio.setGestationDay(0);                          // reset term.

        // recalculate the metabolic rate needed for pregnancy to succeed.
        hominid.bio.setCaloricRequirement(hominid.bio.momentaryCaloricReq(
                                          hominid.bio.getAge(),
                                          hominid.bio.getSex()));
        hominid.bio.setWaterRequirement(hominid.bio.momentaryWaterReq(
                                        hominid.bio.getAge(),
                                        hominid.bio.getSex()));
        hominid.bio.setFatalWeight();   // min weight agent can have and exist.
        hominid.setDriveToDisperse (false);//reset drive to mate outsde cohort.

        father = null;                // return memory to the operating system.
      }
      breeders = null;                // return memory to the operating system.

    }

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method operates on the Hominid through predation.  Most survive,
      a few die*/

  public Hominid predation (Hominid hominid)
  {

    float hiRate = chance_of_predation_hi;
    float loRate = chance_of_predation_lo;
    float dailyRate = loRate + (RNG.nextFloat() * (hiRate - loRate));

    if (RNG.nextFloat() < dailyRate)
    {
      //System.out.println("in LifeExperiences.predation hominid must die");
      hominid.setCauseOfDeath("P");           // death caused by (P) predation.
      hominid.bio.setHominidLifeStatus("D");       // mark hominid as (D) dead.
      hominid.setExtant(false);                              // reduce extancy.
    }

    return hominid;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method sets the current count of incest instances.*/

  public void setIncestCount (int newCount)
  {

    this.incestCount = newCount;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method simulates one's life experiences in the local group.  Here,
      the self interacts with its cohort.*/

  public void socialPsychology (Hominid self, Ecology ecoSys, int h, int v,
                                int epoch)
  {

    // NOTE a baby may respond to these lines.
    // ------------------------------------------------------------------------

    ArrayList <Hominid> localGroup = ecoSys.getCohort (h, v);
    self.kin.updateGroup (localGroup);       // inform oneself of one's cohort.

    // ------------------------------------------------------------------------

    // NOTE a baby will not respond to these lines.
    // ------------------------------------------------------------------------

    if (use_altruism)
    {
      self.altruism (ecoSys);       // if benefactor, engage in social gifting.
    }

    self.alliance (ecoSys, epoch);     // if beneficiary, engage in relocation.

    // doPatrol add code here.

    // ------------------------------------------------------------------------

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method checks for female dead and rereassigns the life phase of
      their infants to "D" dead and "O" orphaned.*/

  public void terminateOrphanInfants(Hominid hominid)
  {

    if (hominid.bio.getHominidLifeStatus().equals("D"))
    {
      // the checks made in order are:
      // if is female and
      // if has children and
      // if last child is not dead and
      // if last child is baby then
      // that child, a (B) baby, is now orphaned and must be terminated.
      if (hominid.bio.getSex().equals("F") &&
          hominid.kin.hasChildren() &&
          hominid.kin.getLastChild().kin.getMother() != null &&
          hominid.kin.getLastChild().bio.getHominidLifeStatus().equals("B"))
      {
        // extinguish infant, set death by orphaning, reduce extancy.
        hominid.kin.getLastChild().bio.setHominidLifeStatus("D");
        hominid.kin.getLastChild().setCauseOfDeath("O");
        hominid.kin.getLastChild().setExtant(false);
      }
    }

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method updates (on a cohort basis) the individual group list(s)
      of the agents identified in the localCohort parameter.*/

  public static void updateCohortAwareness (ArrayList <Hominid> localCohort)
  {

    Hominid member = null;
    Hominid hominid = null;

    int limit = localCohort.size ();
    for (int i = 0; i < limit; i++)
    {
      member = localCohort.get (i);
      for (int j = 0; j < limit; j++)
      {
        hominid = localCohort.get (j);
        if (hominid != member)
          hominid.kin.updateGroup (member);    // FILTER:  only deltas allowed.
                    // updateGroup(m) operates on the individual self.
      }
    }

    member = null;                // return the memory to the operating system.
    hominid = null;               // return the memory to the operating system.

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method captures status data on the existing cohorts and prepares
      a report.*/
  public void diagnosticReport(ArrayList <Hominid> population, int epoch)
  {

    System.out.println("#################################################");
    System.out.println("LifeExperiences.diagnosticReport()" + "\n");
    System.out.println("--------------------------------------------");
    int populationNr = population.size();
    Hominid p = null;
    for (int i = 0; i < populationNr; i++)
    {
      p = population.get(i);                    // listing of hominids in cell.
      reportSubject(p);                  // report contained within this Class.
      p.report(epoch);              // contains most hominid personal features.
      p.kit.kitReport(p);                      // status report of kit on-hand.
      //System.out.println("------------------------");
    }
    System.out.println("Number of hominids reporting: " + populationNr);

    System.out.println("#################################################");

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method displays status data on individual HGs in the cohort passed
      in the parameter.*/
  public void reportSubject(Hominid p)
  {

    System.out.println("#################################################");
    System.out.println("LifeExperiences.reportSubject()");

    System.out.println("Agent ID:       " + p.getAgentID());
    System.out.println("Real location:  " + p.getCurrentLoc().x + "/" +
                                            p.getCurrentLoc().y);
    System.out.println("Age:            " + p.bio.getAge());
    System.out.println("Gender:         " + p.getSexReport());
    System.out.println("Weight (lbs.):  " + p.bio.getCurrentWeight());
    System.out.println("Cause of death: " + p.getCauseOfDeath());

    System.out.println("#################################################");

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method does something.*/

  public void template ()
  {

    // add code here.

  } // End method.
  //***************************************************************************

}
